/*  Copyright (C) 2001  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* ident.c - identify a user */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netinet/in.h>

// Query the remote host's identity.
// Returns result in 'rslt' with length 32.

void get_ident(char *rslt, int sock)
{
  struct sockaddr_in localaddr, remoteaddr, sin_addr;
  struct timeval timeout;
  fd_set *fileset;
  char buf[128], *r, *s;
  int err, fd, len;

  *rslt='\0';	/* start off with an empty result buffer */

  /* set up local/remote addresses */
  len = sizeof(remoteaddr);
  if(getpeername(sock, (struct sockaddr *)&remoteaddr, &len))
    return;
  len = sizeof(localaddr);
  if(getsockname(sock, (struct sockaddr *)&localaddr, &len))
    return;

  /* open socket connection to remote address */
  if((fd=socket(AF_INET, SOCK_STREAM, 0)) == -1)
    return;
  fcntl(fd, F_SETFL, O_NONBLOCK);
  sin_addr.sin_family=AF_INET;
  sin_addr.sin_addr=localaddr.sin_addr;
  sin_addr.sin_port=0;
  if(bind(fd, (struct sockaddr *)&sin_addr, sizeof(sin_addr))) {
    close(fd);
    return;
  }

  /* attempt to do socket connection in progress, timeout 2 seconds */
  sin_addr.sin_family = AF_INET;
  sin_addr.sin_addr = remoteaddr.sin_addr;
  sin_addr.sin_port = htons(113); /* identd port */
  connect(fd, (struct sockaddr *)&sin_addr, sizeof(sin_addr));

  /* Dynamically allocate fd_set with enough padding for 64-bit longs */
  fileset=alloca((fd/64+1)*8);
  memset(fileset, 0, (fd/64+1)*8);
  FD_SET(fd, fileset);

  timeout.tv_sec=2;
  timeout.tv_usec=0;
  if(select(fd+1, 0, fileset, 0, &timeout) < 1) {
    close(fd);
    return;
  }
  len=sizeof(err);
  getsockopt(fd, SOL_SOCKET, SO_ERROR, (char *)&err, &len);
  if(err) {
    close(fd);
    return;
  }

  /* send identification request */
  sprintf(buf, "%d,%d\r\n", ntohs(remoteaddr.sin_port),
          ntohs(localaddr.sin_port));
  write(fd, buf, strlen(buf));

  /* wait for return input */
  timeout.tv_sec=2;
  timeout.tv_usec=0;
  if(select(fd+1, fileset, 0, 0, &timeout) < 1) {
    close(fd);
    return;
  }

  /* read input */
  len=read(fd, buf, 127);
  close(fd);
  if(len <= 2)  /* read error */
    return;
  buf[len]='\0';

  /* parse input string and search for userid */
  for(r=buf,len=0;len<3;len++,r++)
    if(!(r=strchr(r, ':')))
      return;

  if((s=strchr(r, '\n')))
    *s='\0';
  if((s=strchr(r, '\r')))
    *s='\0';

  while(*r == ' ' || *r == '\t')
    r++;
  sprintf(rslt, "%.31s", r);
}
